/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ASMIDE_H
#define __ASMIDE_H

#ifdef G_OS_WIN32
//if we're in Windows, define these directly
#define VERSION "1.0"
#define DATADIR "."
#define DOCDIR "."
#else
//otherwise, we get them from make
#include "../config.h"
#endif

#include <gtk/gtk.h>
#include "file.h"

extern GtkWidget *window, *notebook, *outputwinvbox;
extern GList *open_files;
extern GList *project_widgets;
extern file *currfile;
extern GtkTextTagTable *tagtable;
extern gint winwidth, winheight;
extern gint numrecentfiles;
extern GtkRecentManager *recentfiles;
extern GtkListStore *recent_search_list;
extern GtkTextTag *result_tag, *label_tag, *comment_tag, *directive_tag,
	*opcode_tag, *string_tag, *register_tag;
extern GdkColor resultcolor, labelcolor, commentcolor, dircolor,
	opcodecolor, strcolor, regcolor;
extern gboolean dosyntaxhlt, labelhlt, commenthlt, dirhlt, strhlt, opcodehlt, reghlt;
extern gint numrecentfiles;
extern gint tabwidth;
extern gboolean wordwrap;
extern gboolean autoindent;
extern PangoFontDescription *editorfont;
extern PangoTabArray *tabstops;
extern gunichar colon_unichar, leftparen_unichar, rightparen_unichar, dot_unichar, pound_unichar,
	semicolon_unichar, quote_unichar, comma_unichar, underscore_unichar;
extern gint max_undo_stack_len;
extern gboolean labels_casesensitive;
extern gint output_type;
extern GtkListStore *output_types;
extern GtkListStore *include_dirs;
extern gboolean save_before_build;

typedef enum {
	NONE,
	GREEN,
	RED,
	ORANGE,
	BLUE
} output_color;

typedef enum {
	LOWER,
	UPPER,
	TITLE,
	INVERT
} case_type;

void show_error (gchar*, gchar*);
void show_info (gchar*);
void set_statusbar_text (gchar*);
void update_build_widgets ();
void update_file_widgets (gboolean);
void set_undo_sensitive (gboolean);
void set_redo_sensitive (gboolean);
void update_statusbar (GtkTextIter*);
void update_editor_settings ();
void update_syntaxhlt_settings ();
void clear_assembler_outputwin ();
void update_assembler_outputwin (gboolean);
void append_assembler_output (const gchar*, output_color);
gchar *get_output_type_extension (int);

//name of program - used for window titles, etc.
#define APP_NAME "Wabbitcode"
//number of built-in output types
#define NUM_OUTPUT_TYPES 10

//used to init GdkColor from a 0xRRGGBB-type color definition
#define INIT_COLOR(color, def) { \
		color.red = (def >> 8) & 0xFF00; \
		color.green = def & 0xFF00; \
		color.blue = (def << 8) & 0xFF00; \
	}

#endif
