/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "asmide.h"
#include "file.h"


/*
 * Adds search and replace
 * text to recent searches list
 */

void add_to_recent_searches (const gchar *searchtext, const gchar *replacetext) {
	GtkTreeIter treeiter;
	gchar *testsearchtext, *testreplacetext;

	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (recent_search_list), &treeiter)) {
		do {
			gtk_tree_model_get (GTK_TREE_MODEL (recent_search_list), &treeiter, 0, &testsearchtext,
			                    1, &testreplacetext, -1);

			if (!strcmp (testsearchtext, searchtext)) {
				if (!strcmp (testreplacetext, replacetext) || *replacetext == '\0') {
					g_free (testsearchtext);
					g_free (testreplacetext);
					return;
				} else if (*testreplacetext == '\0' && *replacetext != '\0') {
					gtk_list_store_set (recent_search_list, &treeiter, 1, replacetext, -1);
					g_free (testsearchtext);
					g_free (testreplacetext);
					return;
				}
			}
			g_free (testsearchtext);
			g_free (testreplacetext);

		} while (gtk_tree_model_iter_next (GTK_TREE_MODEL (recent_search_list), &treeiter));
	}

	gtk_list_store_insert_with_values (recent_search_list, NULL, 0, 0, searchtext,
	                                   1, replacetext, -1);
}


/*
 * Searches for text starting from
 * startiter, case sensitivity set by
 * case_sensitive, if found returns TRUE
 * and sets resultstartiter to beginning
 * of found string and resultenditer to
 * end of string
 */

gboolean forward_search (GtkTextIter *startiter, gchar *text,
						 gboolean case_sensitive, GtkTextIter *resultstartiter,
						 GtkTextIter *resultenditer, gboolean use_regexps) {

	/* gtk_text_iter_forward_search still doesn't have a case-sensitivity options,
		so I've got to write an entire friggin' search routine myself... */
	gunichar curr_char, next_char;
	const gchar *tempstr;
	GtkTextIter iter;
	int len;

	if (use_regexps) {
		//strip ? and * from the beginning and end of search strings that are
		// supposed to be regexps, since they don't do anything there and just mean more complex code
		while (TRUE) {
			len = strlen (text);
			if (*text == '*' || *text == '?')
				text++;
			else if (text[len - 1] == '*')
				text[len - 1] = '\0';
			else
				break;
		}
	}

	iter = *startiter;
	curr_char = gtk_text_iter_get_char (&iter);
	if (!curr_char)
		return FALSE;

	while (TRUE) {
		while ((case_sensitive ? curr_char : g_unichar_tolower (curr_char)) !=
		      (case_sensitive ? g_utf8_get_char (text) :
		      g_unichar_tolower (g_utf8_get_char (text)))) {

			gtk_text_iter_forward_char (&iter);
			curr_char = gtk_text_iter_get_char (&iter);
			if (!curr_char)
				return FALSE;
		}

		*resultstartiter = iter;
		tempstr = text;

		while (TRUE) {
			if (use_regexps && *tempstr == '*') {
				//ignore idiotic stuff like "abc*?def"
				if (*(++tempstr) == '?')
					tempstr++;
				next_char = g_utf8_get_char (tempstr);
				if (case_sensitive)
					next_char = g_unichar_tolower (next_char);

				//get the char after the '*' and try to find that
				while ((case_sensitive ? curr_char : g_unichar_tolower (curr_char)) != next_char) {
					gtk_text_iter_forward_char (&iter);
					curr_char = gtk_text_iter_get_char (&iter);
					//if we run into the end of the line before finding it, then no match here!
					if (!curr_char)
						return FALSE;
					if (gtk_text_iter_ends_line (&iter))
						break;
				}

			} else if (use_regexps && *tempstr == '?') {
				//do nothing, this char matches
				// (yes, I know leaving an empty if block like this is ugly,
				//  but I'm too lazy to figure out the convoluted logic I'd
				//  need to add to the next one to leave it out)
			} else if ((case_sensitive ? curr_char : g_unichar_tolower (curr_char)) !=
			           (case_sensitive ? g_utf8_get_char (tempstr) :
			           g_unichar_tolower (g_utf8_get_char (tempstr)))) {
				break;
			}

			gtk_text_iter_forward_char (&iter);
			tempstr = g_utf8_next_char (tempstr);
			if (*tempstr == '\0')
				break;
			curr_char = gtk_text_iter_get_char (&iter);
			if (!curr_char)
				return FALSE;
		}

		if (*tempstr == '\0') {
			*resultenditer = iter;
			return TRUE;
		}

		curr_char = gtk_text_iter_get_char (&iter);
	}
}


/*
 * Shows dialog and goes to line
 */

void callback_goto_line (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *label, *entry;
	gint response, linenum, len, pos;
	GtkTextIter linestartiter, lineenditer;
	const gchar *linenumtext;

	//show the dialog...
	dialog = gtk_dialog_new_with_buttons ("Go to line", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	hbox = gtk_hbox_new (FALSE, 0);
	label = gtk_label_new ("Go to line:");
	entry = gtk_entry_new ();
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 6);
	gtk_widget_show_all (hbox);

	//get the line number entered...
	response = gtk_dialog_run (GTK_DIALOG (dialog));
	linenumtext = gtk_entry_get_text (GTK_ENTRY (entry));

	//make sure there aren't any invalid chars
	len = strlen (linenumtext);
	for (pos = 0; pos < len; pos++) {
		if (!isdigit (linenumtext[pos])) {
			show_info ("This isn't a number!");
			gtk_widget_destroy (dialog);
			return;
		}
	}

	linenum = atoi (linenumtext);
	gtk_widget_destroy (dialog);

	if (response != GTK_RESPONSE_ACCEPT)
		return;

	//and select that line
	gtk_text_buffer_get_iter_at_line (currfile->textbuffer, &linestartiter, linenum - 1);
	lineenditer = linestartiter;
	gtk_text_iter_forward_to_line_end (&lineenditer);
	gtk_text_buffer_select_range (currfile->textbuffer, &linestartiter, &lineenditer);
	gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (currfile->textview), &linestartiter, 0, TRUE, 0, 0);
	gtk_widget_grab_focus (currfile->textview);
}


/*
 * Shows dialog and goes to label
 */

void callback_goto_label (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *label, *entry;
	gint response, pos, len;
	const gchar *labelname, *labelnamesearch;
	GtkTextIter searchiter, selectiter;
	gunichar testchar;
	gboolean foundit;

	//get the name of the label to scroll to
	dialog = gtk_dialog_new_with_buttons ("Go to label", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	hbox = gtk_hbox_new (FALSE, 0);
	label = gtk_label_new ("Go to label:");
	entry = gtk_entry_new ();
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 6);
	gtk_widget_show_all (hbox);

	response = gtk_dialog_run (GTK_DIALOG (dialog));

	if (response != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (dialog);
		return;
	}

	labelname = gtk_entry_get_text (GTK_ENTRY (entry));
	//make sure there aren't any invalid chars
	len = strlen (labelname);
	for (pos = 0; pos < len; pos++) {
		if (!(isalnum (labelname[pos]) || labelname[pos] == '_')) {
			show_info ("Labels can only contain numbers,\nletters, and underscores");
			gtk_widget_destroy (dialog);
			return;
		}
	}

	//now start at the beginning of the file
	foundit = FALSE;
	gtk_text_buffer_get_start_iter (currfile->textbuffer, &searchiter);
	do {
		//and for each line, check if it starts with the same char as the label
		labelnamesearch = labelname;
		testchar = gtk_text_iter_get_char (&searchiter);
		if (g_utf8_get_char (labelnamesearch++) == testchar) {
			gtk_text_iter_forward_char (&searchiter);

			//if it does, check the rest of the line until it hits the end of the label
			while (TRUE) {
				testchar = gtk_text_iter_get_char (&searchiter);
				gtk_text_iter_forward_char (&searchiter);

				if (!testchar || !(g_unichar_isalnum (testchar) || testchar == g_utf8_get_char ("_"))) {
					//magic! you're not expected to understand this
					gtk_text_iter_backward_cursor_position (&searchiter);
					gtk_text_iter_set_line_offset (&searchiter, 0);
					foundit = TRUE;
					break;
				}
				if (g_utf8_get_char (labelnamesearch++) != testchar)
					break;
			}
		}
	} while (!foundit && gtk_text_iter_forward_line (&searchiter));

	gtk_widget_destroy (dialog);

	//if it was found, scroll to the right line
	if (foundit) {
		selectiter = searchiter;
		gtk_text_iter_forward_to_line_end (&selectiter);
		gtk_text_buffer_select_range (currfile->textbuffer, &searchiter, &selectiter);
		gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (currfile->textview), &searchiter, 0, TRUE, 0, 0);
		gtk_widget_grab_focus (currfile->textview);
	} else {
		//otherwise, show a message
		show_info ("Label not found");
	}
}


/*
 * Shows search dialog
 */

void callback_find_changed (GtkComboBox *widget, gpointer dialog) {
	gchar *text;

	text = gtk_combo_box_get_active_text (widget);
	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT, text && *text != '\0');
	if (text)
		g_free (text);
}

void callback_find (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *vbox, *label, *entry, *matchcase, *useregexp;
	GtkTextIter startiter, resultstartiter, resultenditer, bufstartiter;
	gchar *searchtext;
	gint response;
	gboolean wrapping_around = FALSE, result_highlighted = FALSE;

	//show the dialog, etc.
	dialog = gtk_dialog_new_with_buttons ("Find text", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
				GTK_STOCK_FIND, GTK_RESPONSE_ACCEPT, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	hbox = gtk_hbox_new (FALSE, 6);
	label = gtk_label_new ("Search for:");
	entry = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL (recent_search_list), 0);
	g_signal_connect (entry, "changed", G_CALLBACK (callback_find_changed), dialog);
	gtk_entry_set_activates_default (GTK_ENTRY (gtk_bin_get_child (GTK_BIN (entry))), TRUE);
	gtk_combo_box_set_active (GTK_COMBO_BOX (entry), 0);
	matchcase = gtk_check_button_new_with_mnemonic ("Match _case");
	useregexp = gtk_check_button_new_with_mnemonic ("_Use wildcards (* and ?)");
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 2);
	vbox = GTK_DIALOG (dialog)->vbox;
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), matchcase, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), useregexp, TRUE, TRUE, 6);
	gtk_widget_show_all (vbox);

	gtk_text_buffer_get_start_iter (currfile->textbuffer, &bufstartiter);

	response = gtk_dialog_run (GTK_DIALOG (dialog));
	while (response != GTK_RESPONSE_CLOSE && response != GTK_RESPONSE_DELETE_EVENT) {
		//first clear the highlighting for the last search result
		if (result_highlighted) {
			gtk_text_buffer_remove_tag (currfile->textbuffer, result_tag,
				&resultstartiter, &resultenditer);
			result_highlighted = FALSE;
		}

		//now actually search
		if (!wrapping_around)
			searchtext = gtk_combo_box_get_active_text (GTK_COMBO_BOX (entry));

		if (!wrapping_around) {
			//add the search text to the recent searches list
			add_to_recent_searches (searchtext, "");
			//get the start and end points of the search
			gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &startiter,
				gtk_text_buffer_get_selection_bound (currfile->textbuffer));

		} else {
			wrapping_around = FALSE;
		}

		if (forward_search (&startiter, searchtext,
			gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (matchcase)), &resultstartiter,
			&resultenditer, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (useregexp)))) {

			//if it was found, select the result
			gtk_text_buffer_select_range (currfile->textbuffer,
				&resultstartiter, &resultenditer);
			gtk_text_buffer_apply_tag (currfile->textbuffer, result_tag,
				&resultstartiter, &resultenditer);
			result_highlighted = TRUE;

			//and scroll to it
			gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (currfile->textview), &resultstartiter,
				0, FALSE, 0, 0);

		} else {
			//if we've reached the end of the buffer, then wrap around
			if (!gtk_text_iter_equal (&startiter, &bufstartiter)) {
				startiter = bufstartiter;
				wrapping_around = TRUE;
				continue;
			} else {
				//otherwise, it really just doesn't exist
				show_info ("Text not found");
				//so deselect the last result
				gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &startiter,
					gtk_text_buffer_get_insert (currfile->textbuffer));
				gtk_text_buffer_place_cursor (currfile->textbuffer, &startiter);
			}
		}
		g_free (searchtext);
		response = gtk_dialog_run (GTK_DIALOG (dialog));
	}

	if (result_highlighted)
		gtk_text_buffer_remove_tag (currfile->textbuffer, result_tag,
			&resultstartiter, &resultenditer);

	gtk_widget_destroy (dialog);
}


/*
 * Shows replace dialog
 */

#define RESPONSE_REPLACE_ALL 1
#define RESPONSE_REPLACE 2
#define RESPONSE_FIND 3

void callback_replace_activate (GtkEntry *widget, gpointer dialog) {
	gtk_dialog_response (GTK_DIALOG (dialog), RESPONSE_FIND);
}

void callback_replace_changed (GtkComboBox *widget, gpointer dialog) {
	gchar *text;

	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_REPLACE, FALSE);

	text = gtk_combo_box_get_active_text (widget);
	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_FIND, text && *text != '\0');
	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_REPLACE_ALL,
		text && *text != '\0');

	if (text)
		g_free (text);
}

void callback_replace (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *vbox, *label1, *label2, *findentry, *replaceentry, *matchcase, *useregexp;
	GtkRequisition label1size, label2size;
/*	int init_row = 0;
	GtkTreeIter treeiter;
	gchar *tempcol;*/
	GtkTextIter startiter, resultstartiter, resultenditer, bufstartiter;
	gint response, num_replaced;
	gboolean wrapping_around = FALSE, searching_after_replace = FALSE, case_sensitive,
		result_highlighted = FALSE;
	gchar *searchtext, *replacetext, *message_str;

	//show the dialog, etc.
	dialog = gtk_dialog_new_with_buttons ("Find and replace text", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
				"Replace _All", RESPONSE_REPLACE_ALL, "Replace", RESPONSE_REPLACE,
				GTK_STOCK_FIND, RESPONSE_FIND, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), RESPONSE_FIND);
	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_REPLACE, FALSE);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	//find the best entry to put in the find/replace fields
/*	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (recent_search_list), &treeiter)) {
		while (TRUE) {
			gtk_tree_model_get (GTK_TREE_MODEL (recent_search_list), &treeiter, 1, &tempcol, -1);
			if (tempcol && *tempcol != '\0') {
				g_free (tempcol);
				break;
			}
			g_free (tempcol);

			if (gtk_tree_model_iter_next (&treeiter)) {
				init_row++;
			} else {
				init_row = 0;
				break;
			}
		}
	}*/

	vbox = GTK_DIALOG (dialog)->vbox;
	hbox = gtk_hbox_new (FALSE, 6);
	label1 = gtk_label_new ("Search for:");
	findentry = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL (recent_search_list), 0);
	g_signal_connect (findentry, "changed", G_CALLBACK (callback_replace_changed), dialog);
	g_signal_connect (gtk_bin_get_child (GTK_BIN (findentry)), "activate",
		G_CALLBACK (callback_replace_activate), dialog);
	gtk_combo_box_set_active (GTK_COMBO_BOX (findentry), 0);
	gtk_box_pack_start (GTK_BOX (hbox), label1, FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX (hbox), findentry, TRUE, TRUE, 2);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 6);
	hbox = gtk_hbox_new (FALSE, 6);
	label2 = gtk_label_new ("Replace with:");
	replaceentry = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL (recent_search_list), 1);
	g_signal_connect (gtk_bin_get_child (GTK_BIN (replaceentry)), "activate",
		G_CALLBACK (callback_replace_activate), dialog);
	gtk_combo_box_set_active (GTK_COMBO_BOX (replaceentry), 0);
	gtk_box_pack_start (GTK_BOX (hbox), label2, FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX (hbox), replaceentry, TRUE, TRUE, 2);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 6);
	//ugh, nasty stuff to get the labels to line up (I hate packing in GTK)
	gtk_misc_set_alignment (GTK_MISC (label1), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label2), 0, 0);
	gtk_widget_size_request (label1, &label1size);
	gtk_widget_size_request (label2, &label2size);
	if (label1size.width > label2size.width)
		gtk_widget_set_size_request (label2, label1size.width, -1);
	else if (label2size.width > label1size.width)
		gtk_widget_set_size_request (label1, label2size.width, -1);

	matchcase = gtk_check_button_new_with_mnemonic ("Match _case");
	gtk_box_pack_start (GTK_BOX (vbox), matchcase, TRUE, TRUE, 6);
	useregexp = gtk_check_button_new_with_mnemonic ("_Use wildcards (* and ?)");
	gtk_box_pack_start (GTK_BOX (vbox), useregexp, TRUE, TRUE, 6);
	gtk_widget_show_all (vbox);

	response = gtk_dialog_run (GTK_DIALOG (dialog));
	while (response != GTK_RESPONSE_CLOSE && response != GTK_RESPONSE_DELETE_EVENT) {

		//first clear the highlighting for the last search result
		if (result_highlighted) {
			gtk_text_buffer_remove_tag (currfile->textbuffer, result_tag,
				&resultstartiter, &resultenditer);
			result_highlighted = FALSE;
		}

		//then get the search text and the replacement text
		if (!wrapping_around && !searching_after_replace) {
			searchtext = gtk_combo_box_get_active_text (GTK_COMBO_BOX (findentry));
			replacetext = gtk_combo_box_get_active_text (GTK_COMBO_BOX (replaceentry));
		}
		case_sensitive = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (matchcase));

		if (!wrapping_around) {
			//add the search text to the recent searches list
			add_to_recent_searches (searchtext, replacetext);
		}

		if (response == RESPONSE_FIND) {

			/* Find next occurrence */

			if (!wrapping_around) {
				//get the start point for the search
				gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &startiter,
					gtk_text_buffer_get_selection_bound (currfile->textbuffer));
			} else {
				wrapping_around = FALSE;
			}

			if (forward_search (&startiter, searchtext, case_sensitive, &resultstartiter,
				&resultenditer, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (useregexp)))) {

				//if it was found, select the result
				gtk_text_buffer_select_range (currfile->textbuffer,
					&resultstartiter, &resultenditer);
				gtk_text_buffer_apply_tag (currfile->textbuffer, result_tag,
					&resultstartiter, &resultenditer);
				result_highlighted = TRUE;

				//and scroll to it
				gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (currfile->textview), &resultstartiter,
					0, FALSE, 0, 0);

				gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_REPLACE, TRUE);
				searching_after_replace = FALSE;

			} else {
				gtk_text_buffer_get_start_iter (currfile->textbuffer, &bufstartiter);
				//if we've reached the end of the buffer, then wrap around
				if (!gtk_text_iter_equal (&startiter, &bufstartiter)) {
					startiter = bufstartiter;
					wrapping_around = TRUE;
					continue;
				} else {
					//otherwise, it really just doesn't exist
					if (searching_after_replace) {
						show_info ("Done replacing");
						searching_after_replace = FALSE;
					} else {
						show_info ("Text not found");
					}
					gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), RESPONSE_REPLACE, FALSE);
					//so deselect the last result
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &startiter,
						gtk_text_buffer_get_insert (currfile->textbuffer));
					gtk_text_buffer_place_cursor (currfile->textbuffer, &startiter);
				}
			}

		} else if (response == RESPONSE_REPLACE) {

			/* Replace the selected text, and find next */

			gtk_text_buffer_remove_tag (currfile->textbuffer, result_tag,
				&resultstartiter, &resultenditer);
			gtk_text_buffer_begin_user_action (currfile->textbuffer);
			gtk_text_buffer_delete (currfile->textbuffer, &resultstartiter, &resultenditer);
			gtk_text_buffer_insert (currfile->textbuffer, &resultstartiter, replacetext, -1);
			gtk_text_buffer_end_user_action (currfile->textbuffer);
			startiter = resultstartiter;
			searching_after_replace = TRUE;
			response = RESPONSE_FIND;
			continue;

		} else if (response == RESPONSE_REPLACE_ALL) {

			/* Replace all! */

			num_replaced = 0;
			gtk_text_buffer_get_start_iter (currfile->textbuffer, &startiter);
			gtk_text_buffer_begin_user_action (currfile->textbuffer);

			while (forward_search (&startiter, searchtext, case_sensitive, &resultstartiter,
			       &resultenditer, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (useregexp)))) {

				gtk_text_buffer_delete (currfile->textbuffer, &resultstartiter, &resultenditer);
				gtk_text_buffer_insert (currfile->textbuffer, &resultstartiter, replacetext, -1);
				startiter = resultstartiter;
				num_replaced++;
			}

			gtk_text_buffer_end_user_action (currfile->textbuffer);

			message_str = g_strdup_printf ("Replaced %d occurrences", num_replaced);
			show_info (message_str);
			g_free (message_str);
		}

		g_free (searchtext);
		g_free (replacetext);
		response = gtk_dialog_run (GTK_DIALOG (dialog));
	}

	if (result_highlighted)
		gtk_text_buffer_remove_tag (currfile->textbuffer, result_tag,
			&resultstartiter, &resultenditer);

	gtk_widget_destroy (dialog);
}

